from __future__ import annotations

from opengsq.protocol_base import ProtocolBase
from opengsq.protocol_socket import TcpClient


class TeamSpeak3(ProtocolBase):
    """
    This class represents the TeamSpeak 3 Protocol. It provides methods to interact with the TeamSpeak 3 API.
    """

    full_name = "TeamSpeak 3 Protocol"

    def __init__(self, host: str, port: int, voice_port: int, timeout: float = 5):
        """
        Initializes the TeamSpeak3 object with the given parameters.

        :param host: The host of the server.
        :param port: The port of the server.
        :param voice_port: The voice port of the server.
        :param timeout: The timeout for the server connection.
        """
        super().__init__(host, port, timeout)
        self._voice_port = voice_port

    async def get_info(self) -> dict[str, str]:
        """
        Asynchronously retrieves the information of the game server.

        :return: A dictionary containing the information of the game server.
        """
        response = await self.__send_and_receive(b"serverinfo")
        return self.__parse_kvs(response)

    async def get_clients(self) -> list[dict[str, str]]:
        """
        Asynchronously retrieves the list of clients on the game server.

        :return: A list of clients on the game server.
        """
        response = await self.__send_and_receive(b"clientlist")
        return self.__parse_rows(response)

    async def get_channels(self) -> list[dict[str, str]]:
        """
        Asynchronously retrieves the list of channels on the game server.

        :return: A list of channels on the game server.
        """
        response = await self.__send_and_receive(b"channellist -topic")
        return self.__parse_rows(response)

    async def __send_and_receive(self, data: bytes):
        """
        Asynchronously sends the given data to the game server and receives the response.

        :param data: The data to send to the game server.
        :return: The response from the game server.
        """
        with TcpClient() as tcpClient:
            tcpClient.settimeout(self._timeout)
            await tcpClient.connect((self._host, self._port))

            # b'TS3\n\rWelcome to the TeamSpeak 3 ServerQuery interface,
            # type "help" for a list of commands and "help <command>" for information on a specific command.\n\r'
            await tcpClient.recv()

            # b'error id=0 msg=ok\n\r'
            tcpClient.send(f"use port={self._voice_port}\n".encode())
            await tcpClient.recv()

            tcpClient.send(data + b"\x0A")
            response = b""

            while not response.endswith(b"error id=0 msg=ok\n\r"):
                response += await tcpClient.recv()

        # Remove last bytes b'\n\rerror id=0 msg=ok\n\r'
        return response[:-21]

    def __parse_rows(self, response: bytes):
        """
        Parses the rows from the given response.

        :param response: The response to parse rows from.
        :return: A list of dictionaries containing the parsed rows.
        """
        return [self.__parse_kvs(row) for row in response.split(b"|")]

    def __parse_kvs(self, response: bytes) -> dict[str, str]:
        """
        Parses the key-value pairs from the given response.

        :param response: The response to parse key-value pairs from.
        :return: A dictionary containing the parsed key-value pairs.
        """
        kvs = {}

        for kv in response.split(b" "):
            items = kv.split(b"=", 1)
            key = str(items[0], encoding="utf-8", errors="ignore")
            val = (
                str(items[1], encoding="utf-8", errors="ignore")
                if len(items) == 2
                else ""
            )
            kvs[key] = val.replace("\\p", "|").replace("\\s", " ").replace("\\/", "/")

        return kvs


if __name__ == "__main__":
    import asyncio

    async def main_async():
        teamspeak3 = TeamSpeak3(
            host="145.239.200.2", port=10011, voice_port=9987, timeout=5.0
        )
        info = await teamspeak3.get_info()
        print(info)
        clients = await teamspeak3.get_clients()
        print(clients)
        channels = await teamspeak3.get_channels()
        print(channels)

    asyncio.run(main_async())



from enum import IntEnum


class ServerType(IntEnum):
    """
    Indicates the type of server.
    """

    Dedicated = 0x64
    """Dedicated server"""

    Listen = 0x6C
    """Listen server"""

    Proxy = 0x70
    """SourceTV relay (proxy)"""



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Status:
    """
    Represents the server status.
    """

    info: dict[str, str]
    """Server's info. If is_XServerQuery is True, then it includes \\info\\xserverquery\\rules\\xserverquery, else \\basic\\info\\rules\\"""

    players: list[dict[str, str]]
    """Server's players."""

    teams: list[dict[str, str]]
    """Server's teams. Only when is_x_server_query is True."""

    @property
    def is_XServerQuery(self) -> bool:
        """Indicates whether the response is XServerQuery or old response."""
        return "XServerQuery" in self.info



import struct

from opengsq.responses.satisfactory import Status
from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.protocol_socket import UdpClient


class Satisfactory(ProtocolBase):
    """
    This class represents the Satisfactory Protocol. It provides methods to interact with the Satisfactory API.
    """

    full_name = "Satisfactory Protocol"

    async def get_status(self) -> Status:
        """
        Asynchronously retrieves the status of the game server. The status includes the server state, version, and beacon port.
        The server state can be one of the following:
        1 - Idle (no game loaded)
        2 - Currently loading or creating a game
        3 - Currently in game

        :return: A Status object containing the status of the game server.
        """
        # Credit: https://github.com/dopeghoti/SF-Tools/blob/main/Protocol.md

        # Send message id, protocol version
        request = struct.pack("2b", 0, 0) + "opengsq".encode()
        response = await UdpClient.communicate(self, request)
        br = BinaryReader(response)
        header = br.read_byte()

        if header != 1:
            raise InvalidPacketException(
                "Packet header mismatch. Received: {}. Expected: {}.".format(
                    chr(header), chr(1)
                )
            )

        br.read_byte()  # Protocol version
        br.read_bytes(8)  # Request data

        return Status(
            state=br.read_byte(), version=br.read_long(), beacon_port=br.read_short()
        )


if __name__ == "__main__":
    import asyncio

    async def main_async():
        satisfactory = Satisfactory(host="79.136.0.124", port=15777, timeout=5.0)
        status = await satisfactory.get_status()
        print(status)

    asyncio.run(main_async())



from __future__ import annotations

from dataclasses import dataclass
from typing import Union


@dataclass
class Status:
    """
    Represents the server status.
    """

    info: dict[str, str]
    """Server's info."""

    players: list[dict[str, Union[int, str]]]
    """Server's players."""



from __future__ import annotations

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.protocol_socket import UdpClient
from opengsq.responses.ase import Player, Status


class ASE(ProtocolBase):
    """
    This class represents the All-Seeing Eye Protocol. It provides methods to interact with the All-Seeing Eye API.
    """

    full_name = "All-Seeing Eye Protocol"

    _request = b"s"
    _response = b"EYE1"

    async def get_status(self) -> Status:
        """
        Asynchronously retrieves the status of the game server.

        This method sends a request to the server and parses the response to create a Status object.

        :return: The status of the game server.
        """
        response = await UdpClient.communicate(self, self._request)

        br = BinaryReader(response)
        header = br.read_bytes(4)
        InvalidPacketException.throw_if_not_equal(header, self._response)

        return Status(
            game_name=br.read_pascal_string(),
            game_port=int(br.read_pascal_string()),
            hostname=br.read_pascal_string(),
            game_type=br.read_pascal_string(),
            map=br.read_pascal_string(),
            version=br.read_pascal_string(),
            password=br.read_pascal_string() != "0",
            num_players=int(br.read_pascal_string()),
            max_players=int(br.read_pascal_string()),
            rules=self.__parse_rules(br),
            players=self.__parse_players(br),
        )

    def __parse_rules(self, br: BinaryReader) -> dict[str, str]:
        rules = {}

        while not br.is_end():
            key = br.read_pascal_string()

            if not key:
                break

            rules[key] = br.read_pascal_string()

        return rules

    def __parse_players(self, br: BinaryReader) -> list[Player]:
        players: list[Player] = []

        while not br.is_end():
            flags = br.read_byte()
            player = {}

            if flags & 1 == 1:
                player["name"] = br.read_pascal_string()

            if flags & 2 == 2:
                player["team"] = br.read_pascal_string()

            if flags & 4 == 4:
                player["skin"] = br.read_pascal_string()

            if flags & 8 == 8:
                try:
                    player["score"] = int(br.read_pascal_string())
                except ValueError:
                    player["score"] = 0

            if flags & 16 == 16:
                try:
                    player["ping"] = int(br.read_pascal_string())
                except ValueError:
                    player["ping"] = 0

            if flags & 32 == 32:
                try:
                    player["time"] = int(br.read_pascal_string())
                except ValueError:
                    player["time"] = 0

            players.append(Player(**player))

        return players


if __name__ == "__main__":
    import asyncio

    async def main_async():
        # mtasa
        ase = ASE(host="79.137.97.3", port=22126, timeout=10.0)
        status = await ase.get_status()
        print(status)

    asyncio.run(main_async())



from __future__ import annotations
from typing import Optional
from dataclasses import dataclass


@dataclass
class Info:
    """
    Represents the info of a game.
    """

    hostname: str
    """The hostname of the game server."""

    num_players: int
    """The number of players in the game."""

    max_players: int
    """The maximum number of players allowed in the game."""

    game_type: str
    """The type of the game."""

    map: str
    """The current map of the game."""

    rounds_played: int
    """The number of rounds played."""

    rounds_total: int
    """The total number of rounds."""

    teams: list[float]
    """The list of teams."""

    target_score: int
    """The target score."""

    status: str
    """The status of the game."""

    ranked: bool
    """Whether the game is ranked."""

    punk_buster: bool
    """Whether PunkBuster is enabled."""

    password: bool
    """Whether a password is required."""

    uptime: int
    """The uptime of the game server."""

    round_time: int
    """The round time."""

    mod: Optional[str] = None
    """The game mod. This property is optional."""

    ip_port: Optional[str] = None
    """The IP port of the game server. This property is optional."""

    punk_buster_version: Optional[str] = None
    """The version of PunkBuster. This property is optional."""

    join_queue: Optional[bool] = None
    """Whether the join queue is enabled. This property is optional."""

    region: Optional[str] = None
    """The region of the game server. This property is optional."""

    ping_site: Optional[str] = None
    """The ping site of the game server. This property is optional."""

    country: Optional[str] = None
    """The country of the game server. This property is optional."""

    blaze_player_count: Optional[int] = None
    """The number of players in the Blaze game state. This property is optional."""

    blaze_game_state: Optional[str] = None
    """The Blaze game state. This property is optional."""

    quick_match: Optional[bool] = None
    """Whether quick match is enabled. This property is optional."""



from dataclasses import dataclass


@dataclass
class Status:
    """
    Represents the status of a server.
    """

    server_id: int
    """The server ID."""

    server_ip: str
    """The IP address of the server."""

    game_port: int
    """The game port of the server."""

    query_port: int
    """The query port of the server."""

    server_name: str
    """The name of the server."""

    map_name: str
    """The name of the map."""

    game_type: str
    """The type of the game."""

    num_players: int
    """The number of players."""

    max_players: int
    """The maximum number of players."""

    ping: int
    """The ping."""

    flags: int
    """The flags."""

    skill: str
    """The skill level."""



from enum import IntEnum


class Environment(IntEnum):
    """
    Indicates the operating system of the server.
    """

    Linux = 0x6C
    """Linux"""

    Windows = 0x77
    """Windows"""

    Mac = 0x6D
    """Mac"""



from dataclasses import dataclass


@dataclass
class StatusPre17:
    """
    Represents the status of a game for versions prior to 1.7.
    """

    protocol: str
    """The protocol of the game."""

    version: str
    """The version of the game."""

    motd: str
    """The message of the day."""

    num_players: int
    """The number of players in the game."""

    max_players: int
    """The maximum number of players allowed in the game."""



from opengsq.protocols import GameSpy3


class GameSpy4(GameSpy3):
    """
    This class represents the GameSpy Protocol version 4. It inherits from the GameSpy3 class and overrides some of its properties.
    """

    full_name = "GameSpy Protocol version 4"
    challenge_required = True


if __name__ == "__main__":
    import asyncio

    async def main_async():
        gs4 = GameSpy4(host="play.avengetech.me", port=19132, timeout=5.0)
        server = await gs4.get_status()
        print(server)

    asyncio.run(main_async())



import random

from enum import Enum

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import AuthenticationException, InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.protocol_socket import TcpClient


class SourceRcon(ProtocolBase):
    """
    This class represents the Source RCON Protocol. It provides methods to interact with the Source RCON API.
    """

    full_name = "Source RCON Protocol"

    def __init__(self, host: str, port: int = 27015, timeout: float = 5.0):
        """
        Initializes the SourceRcon object with the given parameters.

        :param host: The host of the server.
        :param port: The port of the server.
        :param timeout: The timeout for the server connection.
        """
        super().__init__(host, port, timeout)

        self._tcpClient = None

    def __enter__(self):
        """
        Defines what the context manager should do at the beginning of the block created by the with statement.
        Returns the object to be used in the context of the with statement.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Defines what the context manager should do after its block has been executed (or terminates).
        It can suppress exceptions, if needed.
        """
        self.close()

    def close(self):
        """
        Closes the connection to the server.
        """
        if self._tcpClient:
            self._tcpClient.close()

    async def authenticate(self, password: str):
        """
        Asynchronously authenticates the connection to the server using the given password.

        :param password: The password to authenticate with.
        """

        # Connect
        self._tcpClient = TcpClient()
        self._tcpClient.settimeout(self._timeout)
        await self._tcpClient.connect((self._host, self._port))

        # Send password
        id = random.randrange(4096)
        self._tcpClient.send(
            self.__Packet(
                id, self.__PacketType.SERVERDATA_AUTH.value, password
            ).get_bytes()
        )

        # Receive and parse as Packet
        response_data = await self._tcpClient.recv()
        packet = self.__Packet(response_data)

        # Sometimes it will return a PacketType.SERVERDATA_RESPONSE_VALUE, so receive again
        if packet.type != self.__PacketType.SERVERDATA_AUTH_RESPONSE.value:
            response_data = await self._tcpClient.recv()
            packet = self.__Packet(response_data)

        # Throw exception if not PacketType.SERVERDATA_AUTH_RESPONSE
        if packet.type != self.__PacketType.SERVERDATA_AUTH_RESPONSE.value:
            self._tcpClient.close()
            raise InvalidPacketException(
                "Packet header mismatch. Received: {}. Expected: {}.".format(
                    chr(packet.type),
                    chr(self.__PacketType.SERVERDATA_AUTH_RESPONSE.value),
                )
            )

        # Throw exception if authentication failed
        if packet.id == -1 or packet.id != id:
            self._tcpClient.close()
            raise AuthenticationException("Authentication failed")

    async def send_command(self, command: str) -> str:
        """
        Asynchronously sends a command to the server.

        :param command: The command to send to the server.
        :return: The response from the server.
        """

        # Send the command and a empty command packet
        id = random.randrange(4096)
        dummy_id = id + 1
        self._tcpClient.send(
            self.__Packet(
                id, self.__PacketType.SERVERDATA_EXECCOMMAND.value, command
            ).get_bytes()
        )
        self._tcpClient.send(
            self.__Packet(
                dummy_id, self.__PacketType.SERVERDATA_EXECCOMMAND.value, ""
            ).get_bytes()
        )

        packet_bytes = bytes([])
        response = ""

        while True:
            # Receive
            response_data = await self._tcpClient.recv()

            # Concat to last unused bytes
            packet_bytes += response_data

            # Get the packets and get the unused bytes
            packets, packet_bytes = self.__get_packets(packet_bytes)

            # Loop all packets
            for packet in packets:
                if packet.id == dummy_id:
                    return response

                response += packet.body

    # Handle Multiple-packet Responses
    def __get_packets(self, packet_bytes: bytes):
        """
        Retrieves the packets from the given packet bytes.

        :param packet_bytes: The packet bytes to retrieve packets from.
        :return: A list of packets and any remaining bytes.
        """
        packets: list[SourceRcon.__Packet] = []

        br = BinaryReader(packet_bytes)

        # + 4 to ensure br.ReadInt32() is readable
        while br.stream_position + 4 < len(packet_bytes):
            size = br.read_long()

            if br.stream_position + size > len(packet_bytes):
                return packets, packet_bytes[br.stream_position - 4 :]

            id = br.read_long()
            type = br.read_long()
            body = br.read_string()
            br.read_byte()

            packets.append(self.__Packet(id, type, body))

        return packets, bytes([])

    class __PacketType(Enum):
        SERVERDATA_AUTH = 3
        SERVERDATA_AUTH_RESPONSE = 2
        SERVERDATA_EXECCOMMAND = 2
        SERVERDATA_RESPONSE_VALUE = 0

    class __Packet:
        def __init__(self, *args):
            if len(args) == 3:
                self.id = args[0]
                self.type = args[1]
                self.body = args[2]
            else:
                # Single-packet Responses
                br = BinaryReader(args[0])
                br.read_long()
                self.id = br.read_long()
                self.type = br.read_long()
                self.body = br.read_string()

        def get_bytes(self):
            packet_bytes = self.id.to_bytes(4, byteorder="little")
            packet_bytes += self.type.to_bytes(4, byteorder="little")
            packet_bytes += str.encode(self.body + "\0")
            return len(packet_bytes).to_bytes(4, byteorder="little") + packet_bytes


if __name__ == "__main__":
    import asyncio

    async def main_async():
        with SourceRcon(host="", port=27015, timeout=5.0) as rcon:
            await rcon.authenticate("")
            response = await rcon.send_command("cvarlist")
            print(response)

    asyncio.run(main_async())



from __future__ import annotations

import re
from typing import Union

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.protocol_socket import UdpClient
from opengsq.responses.doom3 import Status


class Doom3(ProtocolBase):
    """
    This class represents the Doom3 Protocol. It provides methods to interact with the Doom3 API.
    """

    full_name = "Doom3 Protocol"

    _player_fields = {
        "doom": ["id", "ping", "rate", "name"],
        "quake4": ["id", "ping", "rate", "name", "clantag"],
        "etqw": ["id", "ping", "name", "clantag_pos", "clantag", "typeflag"],
    }

    async def get_status(self, strip_color=True) -> Status:
        """
        Asynchronously retrieves the status of the game server.

        :param strip_color: A boolean indicating whether to strip color codes from the player names.
        :return: A Status object containing the server information and player list.

        This function sends a request to the game server and processes the response to extract server information and player details. If the 'strip_color' parameter is set to True, color codes in player names are removed. The function returns a Status object which includes a dictionary of server information and a list of players.
        """
        request = b"\xFF\xFFgetInfo\x00ogsq\x00"
        response = await UdpClient.communicate(self, request)

        # Remove the first two 0xFF
        br = BinaryReader(response[2:])
        header = br.read_string()

        if header != "infoResponse":
            raise InvalidPacketException(
                f"Packet header mismatch. Received: {header}. Expected: infoResponse."
            )

        # Read challenge
        br.read_bytes(4)

        if br.read_bytes(4) != b"\xff\xff\xff\xff":
            br.stream_position -= 4

        info = {}

        # Read protocol version
        minor = br.read_short()
        major = br.read_short()
        info["version"] = f"{major}.{minor}"

        # Read packet size
        if br.read_long() != br.remaining_bytes():
            br.stream_position -= 4

        # Key / value pairs, delimited by an empty pair
        while br.remaining_bytes() > 0:
            key = br.read_string().strip()
            val = br.read_string().strip()

            if key == "" and val == "":
                break

            info[key] = Doom3.strip_colors(val) if strip_color else val

        players = []

        stream_position = br.stream_position

        # Try parse the fields
        for mod in self._player_fields.keys():
            try:
                players = self.__parse_player(br, self._player_fields[mod], strip_color)
                break
            except Exception:
                players = []
                br.stream_position = stream_position

        status = Status(info, players)

        return status

    def __parse_player(
        self, br: BinaryReader, fields: list, strip_color: bool
    ) -> list[dict[str, Union[int, str]]]:
        """
        Parses the player information from the BinaryReader object.

        :param br: The BinaryReader object containing the player information.
        :param fields: A list of fields to parse from the BinaryReader object.
        :param strip_color: A boolean indicating whether to strip color codes from the player information.
        :return: A list of dictionaries containing the player information.
        """
        players = []

        def value_by_field(field: str, br: BinaryReader):
            if field == "id" or field == "clantag_pos" or field == "typeflag":
                return br.read_byte()
            elif field == "ping":
                return br.read_short()
            elif field == "rate":
                return br.read_long()

            string = br.read_string()

            return Doom3.strip_colors(string) if strip_color else string

        while True:
            player = {field: value_by_field(field, br) for field in fields}

            if player["id"] == 32:
                break

            players.append(player)

        return players

    @staticmethod
    def strip_colors(text: str) -> str:
        """
        Strips color codes from the given text.

        :param text: The text to strip color codes from.
        :return: The text with color codes stripped.
        """
        return re.compile("\\^(X.{6}|.)").sub("", text)


if __name__ == "__main__":
    import asyncio

    async def main_async():
        # doom3
        doom3 = Doom3(host="66.85.14.240", port=27666, timeout=5.0)
        info = await doom3.get_status()
        print(info)

        # etqw
        doom3 = Doom3(host="178.162.135.83", port=27735, timeout=5.0)
        info = await doom3.get_status()
        print(info)

        # quake4
        doom3 = Doom3(host="88.99.0.7", port=28007, timeout=5.0)
        info = await doom3.get_status()
        print(info)

    asyncio.run(main_async())



from __future__ import annotations

import struct

from opengsq.responses.samp import Player, Status
from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.protocol_socket import Socket, UdpClient


class Samp(ProtocolBase):
    """
    This class represents the San Andreas Multiplayer Protocol. It provides methods to interact with the San Andreas Multiplayer API.
    """

    full_name = "San Andreas Multiplayer Protocol"

    _request_header = b"SAMP"
    _response_header = b"SAMP"

    async def get_status(self) -> Status:
        """
        Asynchronously retrieves the status of the game server.

        :return: A Status object containing the status of the game server.
        """
        br = await self.__send_and_receive(b"i")

        return Status(
            password=br.read_byte() != 0,
            num_players=br.read_short(),
            max_players=br.read_short(),
            server_name=self.__read_string(br, 4),
            game_type=self.__read_string(br, 4),
            language=self.__read_string(br, 4),
        )

    async def get_players(self) -> list[Player]:
        """
        Asynchronously retrieves the players of the game server. The server may not respond when the number of players is greater than 100.

        :return: A list containing the players of the game server.
        """
        br = await self.__send_and_receive(b"d")
        numplayers = br.read_short()
        players = [
            Player(
                id=br.read_byte(),
                name=self.__read_string(br),
                score=br.read_long(),
                ping=br.read_long(),
            )
            for _ in range(numplayers)
        ]
        return players

    async def get_rules(self) -> dict[str, str]:
        """
        Asynchronously retrieves the rules of the game server.

        :return: A dictionary containing the rules of the game server.
        """
        br = await self.__send_and_receive(b"r")
        numrules = br.read_short()

        return dict(
            (self.__read_string(br), self.__read_string(br)) for _ in range(numrules)
        )

    async def __send_and_receive(self, data: bytes):
        """
        Asynchronously sends the given data to the game server and receives the response.

        :param data: The data to send to the game server.
        :return: A BinaryReader object containing the response from the game server.
        """
        # Format the address
        host = await Socket.gethostbyname(self._host)
        packet_header = (
            struct.pack("BBBBH", *map(int, host.split(".") + [self._port])) + data
        )
        request = self._request_header + packet_header

        # Validate the response
        response = await UdpClient.communicate(self, request)
        header = response[: len(self._response_header)]

        if header != self._response_header:
            raise InvalidPacketException(
                f"Packet header mismatch. Received: {header}. Expected: {self._response_header}."
            )

        return BinaryReader(response[len(self._response_header) + len(packet_header) :])

    def __read_string(self, br: BinaryReader, read_offset=1):
        """
        Reads a string from the given BinaryReader object.

        :param br: The BinaryReader object to read the string from.
        :param read_offset: The read offset. Defaults to 1.
        :return: The string read from the BinaryReader object.
        """
        length = br.read_byte() if read_offset == 1 else br.read_long()
        return str(br.read_bytes(length), encoding="utf-8", errors="ignore")


if __name__ == "__main__":
    import asyncio

    async def main_async():
        samp = Samp(host="51.254.178.238", port=7777, timeout=5.0)
        status = await samp.get_status()
        print(status)

        await asyncio.sleep(5)
        players = await samp.get_players()
        print(players)

        await asyncio.sleep(5)
        rules = await samp.get_rules()
        print(rules)

    asyncio.run(main_async())



from __future__ import annotations

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.protocol_socket import UdpClient
from opengsq.responses.gamespy2 import Status


class GameSpy3(ProtocolBase):
    """
    This class represents the GameSpy Protocol version 3. It provides methods to interact with the GameSpy API.
    """

    full_name = "GameSpy Protocol version 3"
    challenge_required = False

    async def get_status(self) -> Status:
        """
        Asynchronously retrieves the status of the game server. The status includes information about the server,
        players, and teams.

        :return: A Status object containing the status of the game server.
        """
        # Connect to remote host
        with UdpClient() as udpClient:
            udpClient.settimeout(self._timeout)
            await udpClient.connect((self._host, self._port))

            request_h = b"\xFE\xFD"
            timestamp = b"\x04\x05\x06\x07"
            challenge = b""

            if self.challenge_required:
                # Packet 1: Initial request - (https://wiki.unrealadmin.org/UT3_query_protocol#Packet_1:_Initial_request)
                udpClient.send(request_h + b"\x09" + timestamp)

                # Packet 2: First response - (https://wiki.unrealadmin.org/UT3_query_protocol#Packet_2:_First_response)
                response = await udpClient.recv()

                if response[0] != 9:
                    raise InvalidPacketException(
                        "Packet header mismatch. Received: {}. Expected: {}.".format(
                            chr(response[0]), chr(9)
                        )
                    )

                # Packet 3: Second request - (http://wiki.unrealadmin.org/UT3_query_protocol#Packet_3:_Second_request)
                challenge = int(response[5:].decode("ascii").strip("\x00"))
                challenge = (
                    b"" if challenge == 0 else challenge.to_bytes(4, "big", signed=True)
                )

            request_data = request_h + b"\x00" + timestamp + challenge
            udpClient.send(request_data + b"\xFF\xFF\xFF\x01")

            # Packet 4: Server information response
            # (http://wiki.unrealadmin.org/UT3_query_protocol#Packet_4:_Server_information_response)
            response = await self.__read(udpClient)

        br = BinaryReader(response)

        info = {}

        while True:
            key = br.read_string()

            if key == "":
                break

            info[key] = br.read_string()

        status = Status(
            info,
            self.__get_dictionaries(br, "player"),
            self.__get_dictionaries(br, "team"),
        )

        return status

    async def __read(self, udpClient: UdpClient) -> bytes:
        packet_count = -1
        payloads = {}

        while packet_count == -1 or len(payloads) > packet_count:
            response = await udpClient.recv()

            br = BinaryReader(response)
            header = br.read_byte()

            if header != 0:
                raise InvalidPacketException(
                    "Packet header mismatch. Received: {}. Expected: {}.".format(
                        chr(header), chr(0)
                    )
                )

            # Skip the timestamp and splitnum
            br.read_bytes(13)

            # The 'numPackets' byte
            num_packets = br.read_byte()

            # The low 7 bits are the packet index (starting at zero)
            number = num_packets & 0x7F

            # The high bit is whether or not this is the last packet
            if num_packets & 0x80:
                # Set packet_count if it is the last packet
                packet_count = number + 1

            # The object id
            # 0 = server kv information
            # 1 = player_   \x00\x01player_\x00\x00 since \x01
            # 2 = team_t    \x00\x02team_t\x00\x00  since \x02
            # etc...
            obj_id = br.read_byte()
            header = b""

            if obj_id >= 1:
                # The object key name
                string = br.read_string()

                # How many times did the value appear in the previous packet
                count = br.read_byte()

                # Set back the packet header if it didn't appear before
                header = (
                    b"\x00" + bytes([obj_id]) + string.encode() + b"\x00\x00"
                    if count == 0
                    else b""
                )

            payload = header + br.read()[:-1]

            # Remove the last trash string on the payload
            payloads[number] = payload[: payload.rfind(b"\x00") + 1]

        response = b"".join(payloads[number] for number in sorted(payloads))

        return response

    def __get_dictionaries(
        self, br: BinaryReader, object_type: str
    ) -> list[dict[str, str]]:
        kvs: list[dict[str, str]] = []

        # Return if BaseStream is end
        if br.is_end():
            return kvs

        # Skip a byte
        br.read_byte()

        # Player/Team index
        i = 0

        while not br.is_end():
            key = br.read_string()

            if key:
                # Skip \x00
                br.read_byte()

                # Remove the trailing "_t"
                key = key.rstrip("t").rstrip("_")

                # Change the key to name
                if key == object_type:
                    key = "name"

                while not br.is_end():
                    value = br.read_string().strip()

                    if value:
                        # Add a Dictionary object if not exists
                        if len(kvs) < i + 1:
                            kvs.append({})

                        kvs[i][key] = value
                        i += 1
                    else:
                        break

                i = 0
            else:
                break

        return kvs


if __name__ == "__main__":
    import asyncio

    async def main_async():
        gs3 = GameSpy3(host="95.172.92.116", port=29900, timeout=5.0)
        server = await gs3.get_status()
        print(server)

    asyncio.run(main_async())



from __future__ import annotations

import re

from opengsq.responses.gamespy1 import Status
from opengsq.binary_reader import BinaryReader
from opengsq.protocol_base import ProtocolBase
from opengsq.protocol_socket import UdpClient


class GameSpy1(ProtocolBase):
    """GameSpy Protocol version 1"""

    full_name = "GameSpy Protocol version 1"

    # Legacy:UT_Server_Query - (https://wiki.beyondunreal.com/Legacy:UT_Server_Query)
    # Query_commands - (https://wiki.beyondunreal.com/XServerQuery#Query_commands)
    class __Request:
        BASIC = b"\\basic\\"
        INFO = b"\\info\\xserverquery"
        RULES = b"\\rules\\xserverquery"
        PLAYERS = b"\\players\\xserverquery"
        STATUS = b"\\status\\xserverquery"
        TEAMS = b"\\teams\\"

    async def get_basic(self) -> dict[str, str]:
        """
        Asynchronously retrieves the basic information of the game server.

        :return: A dictionary containing the basic information of the game server.
        """
        return self.__parse_as_key_values(
            await self.__connect_and_send(self.__Request.BASIC)
        )

    # Server may still response with Legacy version
    async def get_info(self, xserverquery: bool = True) -> dict[str, str]:
        """
        Asynchronously retrieves the information of the current game running on the server.

        :param xserverquery: A boolean indicating whether to use XServerQuery.
        :return: A dictionary containing the information of the current game.
        """
        data = (
            xserverquery
            and self.__Request.INFO
            or self.__Request.INFO.replace(b"xserverquery", b"")
        )
        return self.__parse_as_key_values(await self.__connect_and_send(data))

    async def get_rules(self, xserverquery: bool = True) -> dict[str, str]:
        """
        Asynchronously retrieves the rules of the current game running on the server.

        :param xserverquery: A boolean indicating whether to use XServerQuery.
        :return: A dictionary containing the rules of the current game.
        """
        data = (
            xserverquery
            and self.__Request.RULES
            or self.__Request.RULES.replace(b"xserverquery", b"")
        )
        return self.__parse_as_key_values(await self.__connect_and_send(data))

    async def get_players(self, xserverquery: bool = True) -> list[dict[str, str]]:
        """
        Asynchronously retrieves the information of each player on the server.

        :param xserverquery: A boolean indicating whether to use XServerQuery.
        :return: A list containing the information of each player.
        """
        data = (
            xserverquery
            and self.__Request.PLAYERS
            or self.__Request.PLAYERS.replace(b"xserverquery", b"")
        )
        return self.__parse_as_object(await self.__connect_and_send(data))

    async def get_status(self, xserverquery: bool = True) -> Status:
        """
        Asynchronously retrieves the status of the game server.

        XServerQuery: \\info\\xserverquery\\rules\\xserverquery\\players\\xserverquery
        Old response: \\basic\\info\\rules\\players\\

        If the server uses XServerQuery, he sends you the new information, otherwise he'll give you back the old information.

        :param xserverquery: A boolean indicating whether to use XServerQuery.
        :return: A Status object containing the status of the game server.
        """
        data = (
            xserverquery
            and self.__Request.STATUS
            or self.__Request.STATUS.replace(b"xserverquery", b"")
        )
        br = await self.__connect_and_send(data)

        info = self.__parse_as_key_values(br, is_status=True)
        players = self.__parse_as_object(br, is_player=True)
        teams = [] if br.is_end() else self.__parse_as_object(br)

        return Status(info, players, teams)

    async def get_teams(self) -> list[dict[str, str]]:
        """
        Asynchronously retrieves the information of each team on the server.

        :return: A list containing the information of each team.
        """
        return self.__parse_as_object(
            await self.__connect_and_send(self.__Request.TEAMS)
        )

    # Receive packets and sort it
    async def __get_packets_response(self, udpClient: UdpClient):
        payloads = {}
        packet_count = -1

        # Loop until received all packets
        while packet_count == -1 or len(payloads) < packet_count:
            packet = await udpClient.recv()

            # Get the packet number from query_id
            r = re.compile(rb"\\queryid\\\d+\.(\d+)")
            number, payload = int(r.search(packet).group(1)), r.sub(b"", packet)

            # If it is the last packet, it will contain b'\\final\\' at the end of the response
            if payload.endswith(b"\\final\\"):
                # Save the packet count
                packet_count = number

                # Remove the last b'\\final\\'
                payload = payload[:-7]

            # Save the payload, remove the first byte if it is the first packet
            payloads[number] = payload[1:] if number == 1 else payload

        # Sort the payload and return as bytes
        response = b"".join(payloads[number] for number in sorted(payloads))

        return response

    async def __connect_and_send(self, data) -> BinaryReader:
        # Connect to remote host
        with UdpClient() as udpClient:
            udpClient.settimeout(self._timeout)
            await udpClient.connect((self._host, self._port))

            udpClient.send(data)
            br = BinaryReader(await self.__get_packets_response(udpClient))

        return br

    def __parse_as_key_values(
        self, br: BinaryReader, is_status=False
    ) -> dict[str, str]:
        kv = {}

        # Bind key value
        while br.remaining_bytes() > 0:
            key = br.read_string(b"\\").lower()

            # Check is the end of key_values
            if is_status:
                items = key.split("_")

                if len(items) > 1 and items[1].isdigit():
                    # Read already, so add it back
                    br.prepend_bytes(key.encode() + b"\\")
                    break

            value = br.read_string(b"\\")
            kv[key] = value.strip()

        return kv

    def __parse_as_object(
        self, br: BinaryReader, is_player=False
    ) -> list[dict[str, str]]:
        items, keyhashes, filters = {}, [], []

        while br.remaining_bytes() > 0:
            # Get the key, for example player_1, frags_1, ping_1, etc...
            key = br.read_string(b"\\").lower()

            if is_player and key.startswith("teamname_"):
                # Read already, so add it back
                br.prepend_bytes(key.encode() + b"\\")
                break

            # Extract to name and index, for example name=player, index=1
            matches = re.search(r"(.+?)_(\d+)", key)
            name = matches.group(1)
            name = "player" if name == "playername" else name
            name = "team" if name == "teamname" else name
            index = int(matches.group(2))

            # Get the value, and strip it since some values contain whitespaces
            value = br.read_string(b"\\").strip()

            # Some servers (bf1942) report the same player multiple times, so filter it by keyhash
            if name == "keyhash":
                if value in keyhashes:
                    filters.append(index)
                else:
                    keyhashes.append(value)

            # Create a dict on items if next index appears
            if index not in items:
                items[index] = {}

            # Save
            items[index][name] = value

        # Filter items by filters
        return [v for k, v in items.items() if k not in filters]


if __name__ == "__main__":
    import asyncio

    async def main_async():
        gs1 = GameSpy1(host="51.81.48.224", port=23000, timeout=5.0)  # bfield1942
        # gs1 = GameSpy1(address='139.162.235.20', query_port=7778, timeout=5.0)  # ut
        # gs1 = GameSpy1(address='192.223.24.6', query_port=7778, timeout=5.0)  # ut
        gs1 = GameSpy1(host="141.94.205.35", port=12300, timeout=5.0)  # mohaa
        status = await gs1.get_status()
        print(status)

    asyncio.run(main_async())



from __future__ import annotations


class ExtraDataFlag(int):
    """
    Extra Data Flag (EDF)
    """

    Port = 0x80
    """Port"""

    SteamID = 0x10
    """SteamID"""

    Spectator = 0x40
    """Spectator"""

    Keywords = 0x20
    """Keywords"""

    GameID = 0x01
    """GameID"""

    def __init__(self, flags: int) -> None:
        """
        Initializes the ExtraDataFlag with the given flags.

        :param flags: The flags to initialize the ExtraDataFlag with.
        """
        super().__init__()
        self.flags = flags

    def has_flag(self, flag: int):
        """
        Checks if the ExtraDataFlag has the given flag.

        :param flag: The flag to check.
        :return: True if the ExtraDataFlag has the flag, False otherwise.
        """
        return self.flags & flag



from dataclasses import dataclass
from .partial_info import PartialInfo


@dataclass
class GoldSourceInfo(PartialInfo):
    """
    Obsolete GoldSource Response
    """

    address: str
    """IP address and port of the server."""

    mod: int
    """
    Indicates whether the game is a mod
    0 for Half-Life
    1 for Half-Life mod
    """

    link: str
    """URL to mod website."""

    download_link: str
    """URL to download the mod."""

    version: int
    """Version of mod installed on server."""

    size: int
    """Space (in bytes) the mod takes up."""

    type: int
    """
    Indicates the type of mod:
    0 for single and multiplayer mod
    1 for multiplayer only mod
    """

    dll: int
    """
    Indicates whether mod uses its own DLL:
    0 if it uses the Half-Life DLL
    1 if it uses its own DLL
    """



from __future__ import annotations

from opengsq.responses.battlefield import Info, VersionInfo
from opengsq.binary_reader import BinaryReader
from opengsq.protocol_base import ProtocolBase
from opengsq.protocol_socket import TcpClient


class Battlefield(ProtocolBase):
    """
    This class represents the Battlefield Protocol. It provides methods to interact with the Battlefield API.
    """

    full_name = "Battlefield Protocol"

    _info = b"\x00\x00\x00\x21\x1b\x00\x00\x00\x01\x00\x00\x00\x0a\x00\x00\x00serverInfo\x00"
    _version = (
        b"\x00\x00\x00\x22\x18\x00\x00\x00\x01\x00\x00\x00\x07\x00\x00\x00version\x00"
    )
    _players = b"\x00\x00\x00\x23\x24\x00\x00\x00\x02\x00\x00\x00\x0b\x00\x00\x00listPlayers\x00\x03\x00\x00\x00all\x00"

    async def get_info(self) -> Info:
        """
        Asynchronously retrieves the information of the game server.

        :return: An Info object containing the information of the game server.
        """
        data = await self.__get_data(self._info)

        info = {}
        info["hostname"] = str(data.pop(0)).strip()
        info["num_players"] = int(data.pop(0))
        info["max_players"] = int(data.pop(0))
        info["game_type"] = data.pop(0)
        info["map"] = data.pop(0)
        info["rounds_played"] = int(data.pop(0))
        info["rounds_total"] = int(data.pop(0))
        num_teams = int(data.pop(0))
        info["teams"] = [float(data.pop(0)) for _ in range(num_teams)]
        info["target_score"] = int(data.pop(0))
        info["status"] = data.pop(0)
        info["ranked"] = data.pop(0) == "true"
        info["punk_buster"] = data.pop(0) == "true"
        info["password"] = data.pop(0) == "true"
        info["uptime"] = int(data.pop(0))
        info["round_time"] = int(data.pop(0))

        try:
            if data[0] == "BC2":
                info["mod"] = data.pop(0)
                data.pop(0)

            info["ip_port"] = data.pop(0)
            info["punk_buster_version"] = data.pop(0)
            info["join_queue"] = data.pop(0) == "true"
            info["region"] = data.pop(0)
            info["ping_site"] = data.pop(0)
            info["country"] = data.pop(0)

            try:
                info["blaze_player_count"] = int(data[0])
                info["blaze_game_state"] = data[1]
            except Exception:
                info["quick_match"] = data.pop(0) == "true"
        except Exception:
            pass

        return Info(**info)

    async def get_version(self) -> VersionInfo:
        """
        Asynchronously retrieves the version information of the game server.

        :return: A VersionInfo object containing the version information of the game server.
        """
        data = await self.__get_data(self._version)
        return VersionInfo(data[0], data[1])

    async def get_players(self) -> list[dict[str, str]]:
        """
        Asynchronously retrieves the list of players on the game server.

        :return: A list of dictionaries containing the player information.
        """
        data = await self.__get_data(self._players)
        count = int(data.pop(0))  # field count
        fields, data = data[:count], data[count:]
        numplayers = int(data.pop(0))
        players = []

        for _ in range(numplayers):
            values, data = data[:count], data[count:]
            players.append(dict(zip(fields, values)))

        return players

    async def __get_data(self, request: bytes):
        """
        Asynchronously sends the given request to the game server and receives the response.

        :param request: The request to send to the game server.
        :return: A list containing the response data from the game server.
        """
        response = await TcpClient.communicate(self, request)
        return self.__decode(response)

    def __decode(self, response: bytes):
        """
        Decodes the given response from the game server.

        :param response: The response to decode.
        :return: A list containing the decoded response data.
        """
        br = BinaryReader(response)
        br.read_long()  # header
        br.read_long()  # packet length
        count = br.read_long()  # string count
        data = []

        for _ in range(count):
            br.read_long()  # length of the string
            data.append(br.read_string())

        return data[1:]


if __name__ == "__main__":
    import asyncio

    async def main_async():
        entries = [
            ("91.206.15.69", 48888),  # bfbc2
            ("94.250.199.214", 47200),  # bf3
            ("74.91.124.140", 47200),  # bf4
            ("185.189.255.240", 47600),  # bfh
        ]

        for address, query_port in entries:
            battlefield = Battlefield(address, query_port, timeout=10.0)
            info = await battlefield.get_info()
            print(info)
            version = await battlefield.get_version()
            print(version)
            players = await battlefield.get_players()
            print(players)

    asyncio.run(main_async())



from dataclasses import dataclass


@dataclass
class Status:
    """
    Represents the status response from a server.
    """

    password: bool
    """A value indicating whether a password is required to connect to the server."""

    num_players: int
    """The number of players currently connected to the server."""

    max_players: int
    """The maximum number of players that can connect to the server."""

    server_name: str
    """The name of the server."""

    game_type: str
    """The type of game being played on the server."""

    language: str
    """The language of the server."""



from __future__ import annotations

from dataclasses import dataclass, field

from opengsq.responses.ase.player import Player


@dataclass
class Status:
    """
    Represents the status of a game server.
    """

    game_name: str
    """The name of the game."""

    game_port: int
    """The port number of the game server."""

    hostname: str
    """The hostname of the game server."""

    game_type: str
    """The type of the game."""

    map: str
    """The current map of the game."""

    version: str
    """The version of the game."""

    password: bool
    """Whether a password is required to join the game."""

    num_players: int
    """The number of players currently in the game."""

    max_players: int
    """The maximum number of players allowed in the game."""

    rules: dict[str, str] = field(default_factory=dict)
    """The rules of the game. Defaults to an empty dictionary."""

    players: list[Player] = field(default_factory=list)
    """The players currently in the game. Defaults to an empty list."""



class InvalidPacketException(Exception):
    """Represents errors that occur during application execution when a packet is invalid."""

    def __init__(self, message: str):
        """
        Initializes a new instance of the InvalidPacketException class with a specified error message.

        Args:
            message (str): The message that describes the error.
        """
        super().__init__(message)

    @staticmethod
    def throw_if_not_equal(received, expected):
        """
        Checks if the received value is equal to the expected value.

        Args:
            received: The received value.
            expected: The expected value.

        Raises:
            InvalidPacketException: Thrown when the received value does not match the expected value.
        """
        if isinstance(received, bytes) and isinstance(expected, bytes):
            if received != expected:
                raise InvalidPacketException(
                    InvalidPacketException.get_message(received, expected)
                )
        elif received != expected:
            raise InvalidPacketException(
                InvalidPacketException.get_message(received, expected)
            )

    @staticmethod
    def get_message(received, expected):
        """
        Returns a formatted error message.

        Args:
            received: The received value.
            expected: The expected value.

        Returns:
            str: The formatted error message.
        """
        if isinstance(received, bytes) and isinstance(expected, bytes):
            received_str = " ".join(format(x, "02x") for x in received)
            expected_str = " ".join(format(x, "02x") for x in expected)
        else:
            received_str = str(received)
            expected_str = str(expected)

        return f"Packet header mismatch. Received: {received_str}. Expected: {expected_str}."



from __future__ import annotations

from dataclasses import dataclass
from .player import Player


@dataclass
class Status:
    """
    Represents the status of the server.
    """

    info: dict[str, str]
    """The server information."""

    players: list[Player]
    """The list of players."""
